import jax.numpy as jnp

def get_changepoint_matrix(t : jnp.ndarray, changepoint_t : jnp.array) -> jnp.ndarray:
    """Generates a changepoint matrix based on the time indexes and changepoint time indexes.

    Args:
        t (jnp.ndarray): array with timepoints of shape (n, 1) preferably
        changepoint_t (jnp.array): array with changepoint timepoints of shape (n_changepoints,)

    Returns:
        jnp.ndarray: changepoint matrix - already with discontinuities at changepoits hanlded
    """
    

    expanded_ts = jnp.tile(t.reshape((-1, 1)), (1, len(changepoint_t)))
    A = (expanded_ts >= changepoint_t.reshape((1, -1))).astype(int) * expanded_ts
    cutoff_ts = changepoint_t.reshape((1, -1))
    A = jnp.clip(A - cutoff_ts + 1, 0, None)
    return A


def get_changepoint_timeindexes(
    t: jnp.ndarray, changepoint_interval: int, changepoint_range: float = 0.90
) -> jnp.array:
    """
    Returns an array of time indexes for changepoints based on the given parameters.

    Args:
        t (jnp.ndarray): The array of time values.
        changepoint_interval (int): The interval between changepoints.
        changepoint_range (float, optional): The range of changepoints. Defaults to 0.90.
            If greater than 1, it is interpreted as then number of timepoints.
            If less than zero, it is interpreted as number of timepoints from the end of the time series.

    Returns:
        jnp.array: An array of time indexes for changepoints.
    """
    if changepoint_range < 1 and changepoint_range > 0:
        max_t = t.max() * changepoint_range
    elif changepoint_range >= 1:
        max_t = changepoint_range
    else:
        max_t = t.max() + changepoint_range
    
    changepoint_t = jnp.arange(0, max_t, changepoint_interval)
    return changepoint_t



import itertools
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
import pandas as pd
from numpyro import sample
from numpyro.contrib.control_flow import scan
from numpyro.infer import MCMC, NUTS, Predictive, init_to_mean
from sktime.forecasting.base import BaseForecaster, ForecastingHorizon
from collections import OrderedDict
from prophetverse.engine import MAPInferenceEngine, MCMCInferenceEngine, InferenceEngine
from prophetverse.effects import LinearEffect
from prophetverse.utils.frame_to_array import series_to_tensor
from prophetverse.effects import AbstractEffect
import re
import logging

class BaseBayesianForecaster(BaseForecaster):
    """
    Base class for Bayesian forecasters in hierarchical-prophet.

    Args:
        rng_seed (int): Random number generator seed.
        method (str): Inference method to use. Currently, only "mcmc" is supported.
        num_samples (int): Number of MCMC samples to draw.
        num_warmup (int): Number of warmup steps for MCMC.
        num_chains (int): Number of MCMC chains to run.
        
        *args: Additional positional arguments.
        **kwargs: Additional keyword arguments.
    """

    _tags = {
        "capability:pred_int": True,
        "capability:pred_int:insample": True,
    }

    def __init__(
        self,
        rng_key,
        inference_method,
        mcmc_samples,
        mcmc_warmup,
        mcmc_chains,
        optimizer_steps,
        optimizer_name,
        optimizer_kwargs,
        *args,
        **kwargs,
    ):

        self.rng_key = rng_key
        self.mcmc_samples = mcmc_samples
        self.mcmc_warmup = mcmc_warmup
        self.mcmc_chains = mcmc_chains
        self.inference_method = inference_method
        self.optimizer_steps = optimizer_steps
        self.optimizer_name = optimizer_name
        self.optimizer_kwargs = optimizer_kwargs
        self._sample_sites = set()
        super().__init__(*args, **kwargs)
        self.predictive_samples_ = None

    @property
    def optimizer(self):
        if self.optimizer_name.startswith('optax'):

            from numpyro.optim import optax_to_numpyro
            import optax
            scheduler = optax.cosine_decay_schedule(**self.optimizer_kwargs)

            opt = optax_to_numpyro(
                optax.chain(
                    
                    optax.scale_by_adam(),
                    optax.scale_by_schedule(scheduler),
                    optax.scale(-1.0))
            )
            return opt

        return getattr(numpyro.optim, self.optimizer_name)(**self.optimizer_kwargs)

    def _get_fit_data(self, y, X, fh) -> Dict[str, Any]:
        """
        Get the data required for the Numpyro model.

        This method should be implemented by subclasses.

        Args:
            y (pd.DataFrame): Target variable.
            X (pd.DataFrame): Exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            Dict[str, Any]: Data required for the Numpyro model.
        """
        raise NotImplementedError("Must be implemented by subclass")

    def _get_predict_data(self, X, fh):
        """
        Generate samples from the posterior predictive distribution.

        This method should be implemented by subclasses.

        Args:
            X (pd.DataFrame): Exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            pd.DataFrame: Samples from the posterior predictive distribution.
        """
        raise NotImplementedError("Must be implemented by subclass")

    def model(self, *args, **kwargs):
        """
        Numpyro model.

        This method should be implemented by subclasses.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            Any: Model output.
        """
        raise NotImplementedError("Must be implemented by subclass")

    def _fit(self, y, X, fh):
        """
        Fit the Bayesian forecaster to the training data.

        Args:
            y (pd.DataFrame): Target variable.
            X (pd.DataFrame): Exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            self: The fitted Bayesian forecaster.
        """

        self._set_y_scales(y)
        y  = self._scale_y(y)

        data = self._get_fit_data(y, X, fh)

        self.distributions_ = data.get("distributions", {})

        if self.inference_method == "mcmc":
            self.inference_engine_ = MCMCInferenceEngine(self.model, num_samples=self.mcmc_samples, num_warmup=self.mcmc_warmup, num_chains=self.mcmc_chains, rng_key=self.rng_key)
        elif self.inference_method == "map":
            self.inference_engine_ = MAPInferenceEngine(
                self.model,
                rng_key=self.rng_key,
                optimizer=self.optimizer,
                num_steps=self.optimizer_steps,
            )
        else:
            raise ValueError(f"Unknown method {self.inference_method}")

        self.inference_engine_.infer(**data)
        self.posterior_samples_ = self.inference_engine_.posterior_samples_

        return self

    def _predict(self, fh, X):
        """
        Generate point forecasts for the given forecasting horizon.

        Args:
            fh (ForecastingHorizon): Forecasting horizon.
            X (pd.DataFrame): Exogenous variables.

        Returns:
            pd.DataFrame: Point forecasts for the forecasting horizon.
        """
        predictive_samples = self.predict_samples(fh=fh, X=X)
        self.forecast_samples_ = predictive_samples
        y_pred = predictive_samples.mean(axis=1).to_frame(self._y.columns[0])

        return y_pred

    def predict_all_sites(self, fh, X=None):

        if not isinstance(fh, ForecastingHorizon):
            fh = self._check_fh(fh)

        fh_as_index = self.fh_to_index(fh)

        predict_data = self._get_predict_data(X=X,fh= fh)

        predictive_samples_ = self.inference_engine_.predict(**predict_data)
        out = pd.DataFrame(
            data={
                site: data.mean(axis=0).flatten()
                for site, data in predictive_samples_.items()
            },
            index=self.periodindex_to_multiindex(fh_as_index),
        ).sort_index()
        return self._inv_scale_y(out)

    def predict_samples(
        self,
        fh,
        X=None
):
        """
        Generate samples from the posterior predictive distribution.

        Args:
            X (pd.DataFrame): Exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            pd.DataFrame: Samples from the posterior predictive distribution.
        """
        fh_as_index = self.fh_to_index(fh)

        predict_data = self._get_predict_data(X=X, fh=fh)

        self.predictive_samples_ = self.inference_engine_.predict(**predict_data)

        observation_site = self.predictive_samples_["obs"]
        n_samples = self.predictive_samples_["obs"].shape[0]
        preds = pd.DataFrame(
            data=observation_site.T.reshape((-1, n_samples)),
            columns=list(range(n_samples)),
            index=self.periodindex_to_multiindex(fh_as_index),
        ).sort_index()

        return self._inv_scale_y(preds)

    def _set_y_scales(self, y):
        if y.index.nlevels == 1:
            self._scale = y.abs().max().values[0]
        else:
            self._scale = y.groupby(level=list(range(y.index.nlevels - 1))).agg(lambda x: np.abs(x).max())

    def _scale_y(self, y):
        if y.index.nlevels == 1:
            return y / self._scale
        scale_for_each_obs = self._scale.loc[y.index.droplevel(-1)].values
        return y / scale_for_each_obs

    def _inv_scale_y(self, y):
        if y.index.nlevels == 1:
            return y * self._scale
        scale_for_each_obs = self._scale.loc[y.index.droplevel(-1)].values
        return y * scale_for_each_obs

    def _predict_quantiles(self, fh, X, alpha):
        """
        Generate quantile forecasts for the given forecasting horizon.

        Args:
            fh (ForecastingHorizon): Forecasting horizon.
            X (pd.DataFrame): Exogenous variables.
            alpha (float or List[float]): Quantile(s) to compute.

        Returns:
            pd.DataFrame: Quantile forecasts for the forecasting horizon.
        """
        if isinstance(alpha, float):
            alpha = [alpha]

        forecast_samples_ = self.predict_samples(X=X, fh=fh)

        var_names = self._get_varnames()
        var_name = var_names[0]
        int_idx = pd.MultiIndex.from_product([var_names, alpha])

        quantiles = forecast_samples_.quantile(alpha, axis=1).T
        quantiles.columns = int_idx

        return quantiles

    def periodindex_to_multiindex(self, periodindex: pd.PeriodIndex) -> pd.MultiIndex:
        """
        Convert a PeriodIndex to a MultiIndex.

        Args:
            periodindex (pd.PeriodIndex): PeriodIndex to convert.

        Returns:
            pd.MultiIndex: Converted MultiIndex.
        """
        if self._y.index.nlevels == 1:
            return periodindex

        base_levels = self._y.index.droplevel(-1).unique().tolist()

        # import Iterable
        from collections.abc import Iterable

        # Check if base_levels 0 is a iterable:
        if not isinstance(base_levels[0], tuple):
            base_levels = [(x,) for x in base_levels]

        return pd.Index(
            map(
                lambda x: (*x[0], x[1]),
                itertools.product(base_levels, periodindex),
            ),
            name=self._y.index.names,
        )

    def fh_to_index(self, fh: ForecastingHorizon, y: pd.DataFrame = None):
        """
        Convert a ForecastingHorizon to an index.

        Args:
            fh (ForecastingHorizon): Forecasting horizon.
            y (pd.DataFrame, optional): Target variable. Defaults to None.

        Returns:
            pd.Index: Index corresponding to the forecasting horizon.
        """
        if not isinstance(fh, ForecastingHorizon):
            fh = self._check_fh(fh)

        if y is None:
            y = self._y
        fh_dates = fh.to_absolute_index(cutoff=y.index.get_level_values(-1).max())
        return fh_dates

    @property
    def var_names(self):

        return list(self.distributions_.keys())

    def plot_trace(self, *, figsize=(15,25), compact=True, var_names=None, **kwargs):
        import arviz as az
        

        if var_names is None:
            var_names = self.var_names

        return az.plot_trace(
            az.from_numpyro(self.inference_engine_.mcmc_),
            var_names=var_names,
            filter_vars="regex",
            compact=compact,
            figsize=figsize,
        )

    @property
    def site_names(self) -> list[Any]:
        return list(self.posterior_samples_.keys())


def init_params(distributions) -> dict:
    """
    Initializes the model parameters.
    """

    params = {}

    for coefficient_name, distribution in distributions.items():

        if not isinstance(distribution, OrderedDict):
            params[coefficient_name] = numpyro.sample(coefficient_name, distribution)
        else:
            coefficients = []
            for (
                inner_coefficient_name,
                coefficient_distribution,
            ) in distribution.items():
                coefficients.append(
                    numpyro.sample(inner_coefficient_name, coefficient_distribution)
                )
            params[coefficient_name] = jnp.concatenate(coefficients, axis=0)

    return params


class ExogenousEffectMixin:

    def __init__(self,
                 
                 exogenous_effects: List[AbstractEffect],
                 default_effect = None,
                 **kwargs):

        
        self.exogenous_effects = exogenous_effects
        self.default_effect = default_effect
        super().__init__(**kwargs)
        
        if self.default_effect is None:
            self.default_effect = LinearEffect(
                id="default_exog",
                prior=(dist.Normal, 0, 1),
                effect_mode="additive",
            )

    def _set_custom_effects(self, feature_names):

        effects_and_columns = {}
        columns_with_effects = set()
        exogenous_effects = self.exogenous_effects or {}
        

        for effect in exogenous_effects:

            columns = effect.match_columns(feature_names)

            if columns_with_effects.intersection(columns):
                raise ValueError(
                    "Columns {} are already set".format(
                        columns_with_effects.intersection(columns)
                    )
                )

            if not len(columns):
                logging.warning("No columns match the regex {}".format(effect.regex))

            columns_with_effects = columns_with_effects.union(columns)

            effects_and_columns.update(
                {
                    effect.id: (
                        columns,
                        effect,
                    )
                }
            )

        features_without_effects: set = feature_names.difference(columns_with_effects)

        if len(features_without_effects):

            effects_and_columns.update(
                {
                    "default": (
                        features_without_effects,
                        self.default_effect,
                    )
                }
            )

        self._exogenous_effects_and_columns = effects_and_columns

    def _get_exogenous_data_array(self, X):

        out = {}
        for effect_name, (columns, _) in self._exogenous_effects_and_columns.items():
            # If no columns are found, skip
            if len(columns) == 0:
                continue
            
            if X.index.nlevels == 1:
                array = jnp.array(X[columns].values)
            else:
                array = series_to_tensor(X[columns])

            out[effect_name] = array 

        return out

    @property
    def exogenous_effect_dict(self):
        return {k: v[1] for k, v in self._exogenous_effects_and_columns.items() if len(v[0])}



import pandas as pd
from sktime.transformations.series.fourier import FourierFeatures

def seasonal_transformer(freq="D", yearly_seasonality=False, weekly_seasonality=False) -> pd.DataFrame:
    """
    Transforms the seasonality parameters into a pandas DataFrame.

    Args:
        freq (str, optional): The frequency of the time series. Defaults to "D".
        yearly_seasonality (bool or int, optional): Whether to include yearly seasonality. If bool, 
            the number of Fourier terms is set to 10. If int, the number of Fourier terms is set to 
            the specified value. Defaults to False.
        weekly_seasonality (bool or int, optional): Whether to include weekly seasonality. If bool, 
            the number of Fourier terms is set to 3. If int, the number of Fourier terms is set to 
            the specified value. Defaults to False.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the seasonality parameters.

    Raises:
        ValueError: If yearly_seasonality or weekly_seasonality is not a boolean or an integer.
    """
    sp_list = []
    fourier_term_list = []

    if isinstance(yearly_seasonality, bool):
        yearly_seasonality_num_terms = 10
    elif isinstance(yearly_seasonality, int):
        yearly_seasonality_num_terms = yearly_seasonality
    else:
        raise ValueError("yearly_seasonality must be a boolean or an integer")
    if yearly_seasonality:
        sp_list.append("Y")
        fourier_term_list.append(yearly_seasonality_num_terms)

    if isinstance(weekly_seasonality, bool):
        weekly_seasonality_num_terms = 3
    elif isinstance(weekly_seasonality, int):
        weekly_seasonality_num_terms = weekly_seasonality
    else:
        raise ValueError("weekly_seasonality must be a boolean or an integer")

    if weekly_seasonality:
        sp_list.append("W")
        fourier_term_list.append(weekly_seasonality_num_terms)

    return FourierFeatures(
        sp_list=sp_list, fourier_terms_list=fourier_term_list, freq=freq, keep_original_columns=True
    )



import sys

#  pylint: disable=g-import-not-at-top
from typing import Protocol, TypedDict, Dict


import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist
from ..effects import AbstractEffect

def model(
    t,
    y,
    changepoint_matrix,
    init_trend_params,
    trend_mode,
    data={},
    exogenous_effects: Dict[str, AbstractEffect]={}
    
):
    """
    Defines the Numpyro model.

    Args:
        y (jnp.ndarray): Array of time series data.
        X (jnp.ndarray): Array of exogenous variables.
        t (jnp.ndarray): Array of time values.
    """
    params = init_trend_params()

    # Trend
    changepoint_coefficients = params["changepoint_coefficients"]
    offset = params["offset"]
    capacity = params.get("capacity", None)

    trend = (changepoint_matrix) @ changepoint_coefficients.reshape((-1, 1)) + offset
    if trend_mode == "logistic":
        trend = capacity / (1 + jnp.exp(-trend))


    numpyro.deterministic("trend_", trend)

    mean = trend
    # Exogenous effects
    if exogenous_effects is not None:

        for key, exog_effect in exogenous_effects.items():

            exog_data = data[key]
            effect = exog_effect(trend=trend, data=exog_data)
            numpyro.deterministic(key, effect)
            mean += effect

    noise_scale = params["std_observation"] 

    with numpyro.plate("data", len(mean), dim=-2) as time_plate:
        numpyro.sample(
            "obs",
            dist.Normal(mean.reshape((-1, 1)), noise_scale),
            obs=y,
        )



import re

import numpy as np
import pandas as pd
from sktime.transformations.base import BaseTransformer
from sktime.transformations.compose import TransformerPipeline


class ExpandColumnPerLevel(BaseTransformer):
    """
    Transformer that expands columns based on a regular expression pattern per level of a multi-level index.
    If there are M colums matching the pattern and N distinct multi-level indexes, then the transformer will
    create M * N new columns, setting the value to zero to each of M * (N-1) columns for a given index j if the current index
    i /= j.

    Parameters:
    - columns_regex (list): A list of regular expression patterns to match the columns to be expanded.

    Attributes:
    - matched_columns_ (set): A set of columns that match the regular expression patterns.
    - new_columns_ (dict): A dictionary mapping the new column names to the original column names.

    Methods:
    - fit(X, y=None): Fit the transformer to the input data.
    - transform(X, y=None): Transform the input data by expanding the columns.
    """

    def __init__(self, columns_regex):
        self.columns_regex = columns_regex
        super(ExpandColumnPerLevel, self).__init__()

    def fit(self, X, y=None):
        """
        Fit the transformer to the input data.

        Parameters:
        - X (pd.DataFrame): The input data.
        - y (pd.Series or None): The target variable. Not used in this transformer.

        Returns:
        - self (ExpandColumnPerLevel): The fitted transformer object.
        """
        regex_patterns = [re.compile(pattern) for pattern in self.columns_regex]
        self.matched_columns_ = set(
            [
                col
                for col in X.columns
                for pattern in regex_patterns
                if pattern.match(col)
            ]
        )

        self.new_columns_ = {}
        # Ensure identifiers are tuples
        self.series_identifiers_ = [
            idx if isinstance(idx, tuple) else (idx,)
            for idx in X.index.droplevel(-1).unique()
        ]
        for identifier in self.series_identifiers_:
            for col in self.matched_columns_:
                self.new_columns_[self.get_col_name(col, identifier)] = col
        return self

    @classmethod
    def get_col_name(cls, column, identifier):
        """
        Generate the new column name based on the original column name and the identifier.

        Parameters:
        - column (str): The original column name.
        - identifier (tuple): The identifier for the column.

        Returns:
        - new_col_name (str): The new column name.
        """
        return f"{column}_dup_{'-'.join(identifier)}"

    def transform(self, X, y=None):
        """
        Transform the input data by expanding the columns.

        Parameters:
        - X (pd.DataFrame): The input data.
        - y (pd.Series or None): The target variable. Not used in this transformer.

        Returns:
        - X_transformed (pd.DataFrame): The transformed data with expanded columns.
        """
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("Input must have a multi-level index")
        X_original = X.copy()
        X = X.copy()

        regex_patterns = [re.compile(pattern) for pattern in self.columns_regex]
        matched_columns = self.matched_columns_

        # Ensure identifiers are tuples
        series_identifiers = [
            idx if isinstance(idx, tuple) else (idx,)
            for idx in X.index.droplevel(-1).unique()
        ]

        def get_col_name(column, identifier):
            return f"{column}_dup_{'-'.join(identifier)}"

        for match in matched_columns:
            for identifier in series_identifiers:
                X[get_col_name(match, identifier)] = np.float64(0)

        self.new_columns_ = {}
        for identifier in series_identifiers:

            for column in matched_columns:
                self.new_columns_[get_col_name(column, identifier)] = column
                X.loc[identifier, get_col_name(column, identifier)] = X_original.loc[
                    identifier, column
                ].values.flatten()
                X[get_col_name(column, identifier)] = X[
                    get_col_name(column, identifier)
                ].fillna(0)

        return X.drop(columns=matched_columns)



import numpy as np


def suggest_logistic_rate_and_offset(
    t: np.ndarray, y: np.ndarray, capacities: float or np.ndarray
):
    """
    Suggests the logistic rate and offset based on the given time series data.

    Parameters:
        t (ndarray): The time values of the time series data.
        y (ndarray): The observed values of the time series data.
        capacities (float or ndarray): The capacity or capacities of the time series data.

    Returns:
        m (ndarray): The suggested offset.
        k (ndarray): The suggested logistic rate.

    """

    if y.ndim == 1:
        y = y.reshape(1, -1)
    elif y.ndim == 3:
        # Shape here would be (n_series, n_samples, 1)
        y = y.squeeze()
    if t.ndim == 1:
        t = t.reshape(1, -1)
    elif t.ndim == 3:
        # Shape here would be (n_series, n_samples, 1)
        t = t.squeeze()

    i0, i1 = t.argmin(axis=1), t.argmax(axis=1)
    t0, t1 = t[:, i0].flatten(), t[:, i1].flatten()
    T = t0 - t1
    y0, y1 = y[:, i0].flatten(), y[:, i1].flatten()

    r0 = capacities / y0
    r1 = capacities / y1

    L0 = np.log(r0 - 1)
    L1 = np.log(r1 - 1)

    
    k = (L1 - L0) / T
    m = - (L1 + k * t1)

    return k, m



from typing import Callable
import numpyro
from numpyro.infer.initialization import init_to_mean
from numpyro.infer import SVI, TraceEnum_ELBO, init_to_value, Trace_ELBO, MCMC, NUTS, Predictive
from numpyro.infer.autoguide import AutoDelta
import jax


class InferenceEngine:
    """Class representing an inference engine for a given model.

    Args:
        model (Callable): The model to be used for inference.
        rng_key (Optional[jax.random.PRNGKey]): The random number generator key. 
            If not provided, a default key with value 0 will be used.

    Attributes:
        model (Callable): The model used for inference.
        rng_key (jax.random.PRNGKey): The random number generator key.

    """

    def __init__(self, model: Callable, rng_key=None):
        self.model = model
        if rng_key is None:
            rng_key = jax.random.PRNGKey(0)
        self.rng_key = rng_key

    def infer(self, **kwargs): 
        """Performs inference using the specified model.

        Args:
            **kwargs: Additional keyword arguments to be passed to the model.

        Returns:
            The result of the inference.

        """
        ...

    def predict(self, **kwargs): 
        """Generates predictions using the specified model.

        Args:
            **kwargs: Additional keyword arguments to be passed to the model.

        Returns:
            The predictions generated by the model.

        """
        ...


class MAPInferenceEngine(InferenceEngine):
    """
    Maximum a Posteriori (MAP) Inference Engine.

    This class performs MAP inference using Stochastic Variational Inference (SVI) with AutoDelta guide.
    It provides methods for inference and prediction.

    Args:
        model (Callable): The probabilistic model to perform inference on.
        optimizer (numpyro.optim._NumPyroOptim, optional): The optimizer to use for SVI. Defaults to None.
        num_steps (int, optional): The number of optimization steps to perform. Defaults to 10000.
        rng_key (jax.random.PRNGKey, optional): The random number generator key. Defaults to None.
    """

    def __init__(
        self,
        model: Callable,
        optimizer: numpyro.optim._NumPyroOptim = None,
        num_steps=10000,
        rng_key=None,
    ):
        if optimizer is None:
            optimizer = numpyro.optim.Adam(step_size=0.001)
        self.optimizer = optimizer
        self.num_steps = num_steps
        super().__init__(model, rng_key)

    def infer(self, **kwargs):
        """
        Perform MAP inference.

        Args:
            **kwargs: Additional keyword arguments to be passed to the model.

        Returns:
            self: The updated MAPInferenceEngine object.
        """
        self.guide_ = AutoDelta(self.model, init_loc_fn=init_to_mean())
        self.svi_ = SVI(self.model, self.guide_, self.optimizer, loss=Trace_ELBO())
        self.run_results_ = self.svi_.run(
            rng_key=self.rng_key, num_steps=self.num_steps, **kwargs
        )
        self.posterior_samples_ = self.guide_.sample_posterior(self.rng_key, params=self.run_results_.params, **kwargs)
        return self

    def predict(self, **kwargs):
        """
        Generate predictions using the trained model.

        Args:
            **kwargs: Additional keyword arguments to be passed to the model.

        Returns:
            self.samples_: The predicted samples generated by the model.
        """
        predictive = numpyro.infer.Predictive(
            self.model,
            params=self.run_results_.params,
            guide=self.guide_,
            #posterior_samples=self.posterior_samples_,
            num_samples=1000,
        )
        self.samples_ = predictive(
            rng_key=self.rng_key,
            **kwargs
        )
        return self.samples_


class MCMCInferenceEngine(InferenceEngine):
    """
    MCMCInferenceEngine is a class that performs MCMC (Markov Chain Monte Carlo) inference
    for a given model.

    Args:
        model (Callable): The model function to perform inference on.
        num_samples (int): The number of MCMC samples to draw.
        num_warmup (int): The number of warmup samples to discard.
        num_chains (int): The number of MCMC chains to run in parallel.
        dense_mass (bool): Whether to use dense mass matrix for NUTS sampler.
        rng_key (Optional): The random number generator key.

    Attributes:
        num_samples (int): The number of MCMC samples to draw.
        num_warmup (int): The number of warmup samples to discard.
        num_chains (int): The number of MCMC chains to run in parallel.
        dense_mass (bool): Whether to use dense mass matrix for NUTS sampler.
        mcmc_ (MCMC): The MCMC object used for inference.
        posterior_samples_ (Dict[str, np.ndarray]): The posterior samples obtained from MCMC.
        samples_predictive_ (Dict[str, np.ndarray]): The predictive samples obtained from MCMC.
        samples_ (Dict[str, np.ndarray]): The MCMC samples obtained from MCMC.

    """

    def __init__(
        self,
        model: Callable,
        num_samples=1000,
        num_warmup=200,
        num_chains=1,
        dense_mass=False,
        rng_key=None,
    ):
        self.num_samples = num_samples
        self.num_warmup = num_warmup
        self.num_chains = num_chains
        self.dense_mass = dense_mass
        super().__init__(model, rng_key)

    def infer(self, **kwargs):
        """
        Run MCMC inference.

        Args:
            **kwargs: Additional keyword arguments to be passed to the MCMC run method.

        Returns:
            self: The MCMCInferenceEngine object.

        """
        self.mcmc_ = MCMC(
            NUTS(self.model, dense_mass=self.dense_mass, init_strategy=init_to_mean()),
            num_samples=self.num_samples,
            num_warmup=self.num_warmup,
        )
        self.mcmc_.run(self.rng_key, **kwargs)
        self.posterior_samples_ = self.mcmc_.get_samples()
        return self

    def predict(self, **kwargs):
        """
        Generate predictive samples.

        Args:
            **kwargs: Additional keyword arguments to be passed to the Predictive method.

        Returns:
            Dict[str, np.ndarray]: The predictive samples.

        """
        sites = set(self.posterior_samples_.keys()).union(["obs"])
        predictive = Predictive(self.model, self.posterior_samples_, return_sites=sites)

        self.samples_predictive_ = predictive(self.rng_key, **kwargs)
        self.samples_ = self.mcmc_.get_samples()
        return self.samples_predictive_



import re
from operator import le
from typing import Tuple

import numpy as np
import pandas as pd
from jax import numpy as jnp
from sktime.transformations.hierarchical.reconcile import _get_s_matrix

NANOSECONDS_TO_SECONDS = 1000 * 1000 * 1000


def get_bottom_series_idx(y):
    """
    Get the index of the bottom series in a hierarchical time series.

    Parameters:
    y (pd.DataFrame): The hierarchical time series.

    Returns:
    pd.Index: The index of the bottom series.
    """
    return _get_s_matrix(y).columns


def get_multiindex_loc(df, idx_values):
    """
    Returns the .loc for the first M levels of a multi-indexed DataFrame for given index values.

    Parameters:
    df (pd.DataFrame): The DataFrame with a multi-index.
    idx_values (tuple or list): Index values for the first M levels.

    Returns:
    pd.DataFrame: The selected data from the DataFrame.
    """
    nlevels = df.index.nlevels

    if isinstance(idx_values[0], tuple):
        nlevels_to_loc = len(idx_values[0])
    else:
        nlevels_to_loc = 1

    levels_to_ignore = list(range(-1, -1 - (nlevels - nlevels_to_loc), -1))

    mask = df.index.droplevel(levels_to_ignore).isin(idx_values)
    return df.loc[mask]


def loc_bottom_series(y):
    """
    Get the bottom series in a hierarchical time series.

    Parameters:
    y (pd.DataFrame): The hierarchical time series.

    Returns:
    pd.DataFrame: The bottom series.
    """
    return get_multiindex_loc(y, get_bottom_series_idx(y))


def iterate_all_series(y):
    """
    Iterate over all series in a hierarchical time series.

    Parameters:
    y (pd.DataFrame): The hierarchical time series.

    Yields:
    tuple: A tuple containing the index and the corresponding series.
    """
    series = _get_s_matrix(y).index
    for idx in series:
        yield idx, y.loc[idx]


def convert_index_to_days_since_epoch(idx: pd.Index) -> np.array:
    """
    Convert a pandas Index to days since epoch.

    Parameters:
    idx (pd.Index): The pandas Index.

    Returns:
    np.ndarray: The converted array of days since epoch.
    """
    t = idx
    if isinstance(t, pd.PeriodIndex):
        t = t.to_timestamp()

    return t.to_numpy(dtype=np.int64) // NANOSECONDS_TO_SECONDS / (3600 * 24.0)


def series_to_tensor(y):
    """
    Convert all series of a hierarchical time series to a JAX tensor.

    Parameters:
    y (pd.DataFrame): The hierarchical time series.

    Returns:
    jnp.ndarray: The JAX tensor representing all series.
    """
    names = []
    array = []
    series_len = None
    for idx, series in iterate_all_series(y):
        if series_len is None:
            series_len = len(series)
        if len(series) != series_len:
            raise ValueError("Series {} has length {}, but expected {}".format(idx, len(series), series_len))
        
        names.append(idx)
        array.append(series.values.reshape((-1, len(y.columns))))
    return jnp.array(array)


def extract_timetensor_from_dataframe(df: pd.DataFrame) -> jnp.array:
    """
    Extract the time array from a pandas DataFrame.

    Parameters:
    df (pd.DataFrame): The DataFrame.

    Returns:
    """

    return series_to_tensor(
        pd.DataFrame(
            index=df.index,
            data={
                "t": convert_index_to_days_since_epoch(df.index.get_level_values(-1))
            },
        )
    )

def convert_dataframe_to_tensors(df: pd.DataFrame) -> Tuple[jnp.array, jnp.array]:
    """
    Convert a pandas DataFrame to JAX tensors.

    Parameters:
    df (pd.DataFrame): The DataFrame.

    Returns:
    tuple: A tuple containing the time arrays and the DataFrame arrays as JAX tensors.
    """
    t_arrays = extract_timetensor_from_dataframe(df)

    df_as_arrays = series_to_tensor(df)

    return t_arrays, df_as_arrays



import sys

#  pylint: disable=g-import-not-at-top
from typing import Protocol, TypedDict, Dict


import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist
from prophetverse.effects import AbstractEffect


def model(
    t,
    y,
    changepoint_matrix,
    init_trend_params,
    trend_mode,
    data={},
    exogenous_effects: Dict[str, AbstractEffect] = {},
    noise_scale=0.05,
    correlation_matrix_concentration=1.0,
):
    """
    Defines the Numpyro model.

    Args:
        y (jnp.ndarray): Array of time series data.
        X (jnp.ndarray): Array of exogenous variables.
        t (jnp.ndarray): Array of time values.
    """
    params = init_trend_params()

    # Trend
    changepoint_coefficients = params["changepoint_coefficients"]
    offset = params["offset"]
    capacity = params.get("capacity", None)

    trend = (changepoint_matrix) @ changepoint_coefficients.reshape(
        (1, -1, 1)
    ) + offset.reshape((-1, 1, 1))
    if trend_mode == "logistic":
        trend = (capacity.reshape((-1, 1, 1))) / (1 + jnp.exp(-trend))

    numpyro.deterministic("trend_", trend)

    mean = trend
    # Exogenous effects
    if exogenous_effects is not None:

        for key, exog_effect in exogenous_effects.items():

            exog_data = data[key]
            effect = exog_effect(trend=trend, data=exog_data)
            numpyro.deterministic(key, effect)
            mean += effect

    std_observation = numpyro.sample(
        "std_observation", dist.HalfNormal(jnp.array([noise_scale] * mean.shape[0]))
    )

    if y is not None:
        y = y.squeeze(-1).T
        
    if correlation_matrix_concentration is None:

        with numpyro.plate("time", mean.shape[-1], dim=-2):
            numpyro.sample(
                "obs",
                dist.Normal(mean.squeeze(-1).T, std_observation),
                obs=y
            )

    else:
        correlation_matrix = numpyro.sample(
            "corr_matrix",
            dist.LKJCholesky(
                mean.shape[0],
                concentration=correlation_matrix_concentration,
            ),
        )

        cov_mat = jnp.diag(std_observation) @ correlation_matrix @ jnp.diag(std_observation)

        cov_mat = jnp.tile(jnp.expand_dims(cov_mat, axis=0), (mean.shape[1], 1, 1))

        with numpyro.plate("time", mean.shape[-1], dim=-2):
            numpyro.sample(
                "obs",
                dist.MultivariateNormal(mean.squeeze(-1).T, scale_tril=cov_mat),
                obs=y,
            )



import itertools
import logging
import re
import functools
from typing import Dict, List, Tuple, Iterable

import jax
import jax.numpy as jnp
import numpy as np
import numpyro
from typing import Callable
import pandas as pd
from jax import lax, random
from collections import OrderedDict
from numpyro import distributions as dist
from numpyro.infer import MCMC, NUTS, Predictive
from sktime.forecasting.base import BaseForecaster, ForecastingHorizon
from sktime.transformations.base import BaseTransformer
from sktime.transformations.hierarchical.aggregate import Aggregator
from sktime.transformations.hierarchical.reconcile import _get_s_matrix


from prophetverse.utils.frame_to_array import (
    convert_index_to_days_since_epoch,
    extract_timetensor_from_dataframe,
    loc_bottom_series,
    series_to_tensor,
)
from prophetverse.models.multivariate_model.multiindex import reindex_time_series
from prophetverse.utils.logistic import suggest_logistic_rate_and_offset
from prophetverse.sktime.base import (
    BaseBayesianForecaster,
    ExogenousEffectMixin,
    init_params,
)


from ._expand_column_per_level import (
    ExpandColumnPerLevel,
)

from prophetverse.changepoint import (
    get_changepoint_matrix,
    get_changepoint_timeindexes,
)

from prophetverse.models.multivariate_model._model import model
from prophetverse.effects import LinearEffect, LinearHeterogenousPriorsEffect

logger = logging.getLogger("sktime-numpyro")


class HierarchicalProphet(ExogenousEffectMixin, BaseBayesianForecaster):
    """A class that represents a Bayesian hierarchical time series forecasting model based on the Prophet algorithm.

    This class forecasts all series in a hierarchy at once, using a MultivariateNormal as the likelihood function, and
    LKJ priors for the correlation matrix.

    This class may be interesting if you want to fit shared coefficients across series. By default, all coefficients are
    obtained exclusively for each series, but this can be changed through the `shared_coefficients` parameter.

    Args:
        changepoint_interval (int): The number of points between each potential changepoint.
        changepoint_range (float): Proportion of the history in which trend changepoints will be estimated.
                                   If a float between 0 and 1, the range will be that proportion of the history.
                                   If an int, the range will be that number of points. A negative int indicates number of points
                                   counting from the end of the history.
        changepoint_prior_scale (float): Parameter controlling the flexibility of the automatic changepoint selection.
        capacity_prior_scale (float): Scale parameter for the capacity prior. Defaults to 0.2.
        capacity_prior_loc (float): Location parameter for the capacity prior. Defaults to 1.1.
        trend (str): Type of trend. Either "linear" or "logistic". Defaults to "linear".
        feature_transformer (BaseTransformer or None): A transformer to preprocess the exogenous features. Defaults to None.
        exogenous_effects (List[AbstractEffect]): A list defining the exogenous effects to be used in the model.
        default_effect (AbstractEffect): The default effect to be used when no effect is specified for a variable.
        shared_features (list): List of shared features across series. Defaults to an empty list.
        mcmc_samples (int): Number of MCMC samples to draw. Defaults to 2000.
        mcmc_warmup (int): Number of warmup steps for MCMC. Defaults to 200.
        mcmc_chains (int): Number of MCMC chains. Defaults to 4.
        inference_method (str): Inference method to use. Either "map" or "mcmc". Defaults to "map".
        optimizer_name (str): Name of the optimizer to use. Defaults to "Adam".
        optimizer_kwargs (dict): Additional keyword arguments for the optimizer. Defaults to {"step_size": 1e-4}.
        optimizer_steps (int): Number of optimization steps. Defaults to 100_000.
        noise_scale (float): Scale parameter for the noise. Defaults to 0.05.
        correlation_matrix_concentration (float): Concentration parameter for the correlation matrix. Defaults to 1.0.
        rng_key (jax.random.PRNGKey): Random number generator key. Defaults to random.PRNGKey(24).
    """

    _tags = {
        "scitype:y": "univariate",  # which y are fine? univariate/multivariate/both
        "ignores-exogeneous-X": False,  # does estimator ignore the exogeneous X?
        "handles-missing-data": False,  # can estimator handle missing data?
        "y_inner_mtype": [
            "pd.Series",
            "pd_multiindex_hier",
            "pd-multiindex",
        ],
        "X_inner_mtype": [
            "pd_multiindex_hier",
            "pd-multiindex",
        ],  # which types do _fit, _predict, assume for X?
        "requires-fh-in-fit": False,  # is forecasting horizon already required in fit?
        "X-y-must-have-same-index": False,  # can estimator handle different X/y index?
        "enforce_index_type": None,  # index type that needs to be enforced in X/y
        "capability:pred_int": False,  # does forecaster implement proba forecasts?
        "fit_is_empty": False,
        "capability:pred_int": True,
        "capability:pred_int:insample": True,
    }

    def __init__(
        self,
        changepoint_interval=25,
        changepoint_range=None,
        changepoint_prior_scale=0.1,
        capacity_prior_scale=0.2,
        capacity_prior_loc=1.1,
        trend="linear",
        feature_transformer: BaseTransformer = None,
        exogenous_effects=None,
        default_effect=None,
        shared_features=[],
        mcmc_samples=2000,
        mcmc_warmup=200,
        mcmc_chains=4,
        inference_method="map",
        optimizer_name="Adam",
        optimizer_kwargs={"step_size": 1e-4},
        optimizer_steps=100_000,
        noise_scale=0.05,
        correlation_matrix_concentration=1.0,
        rng_key=random.PRNGKey(24),
    ):

        self.changepoint_interval = changepoint_interval
        self.changepoint_range = changepoint_range
        self.changepoint_prior_scale = changepoint_prior_scale
        self.noise_scale = noise_scale
        self.capacity_prior_scale = capacity_prior_scale
        self.capacity_prior_loc = capacity_prior_loc
        self.trend = trend
        self.shared_features = shared_features
        self.feature_transformer = feature_transformer
        self.correlation_matrix_concentration = correlation_matrix_concentration

        super().__init__(
            rng_key=rng_key,
            inference_method=inference_method,
            optimizer_name=optimizer_name,
            optimizer_kwargs=optimizer_kwargs,
            optimizer_steps=optimizer_steps,
            mcmc_samples=mcmc_samples,
            mcmc_warmup=mcmc_warmup,
            mcmc_chains=mcmc_chains,
            default_effect=default_effect,
            exogenous_effects=exogenous_effects,
        )

        self.model = model
        self._validate_hyperparams()

    def _validate_hyperparams(self):
        """
        Validate the hyperparameters of the HierarchicalProphet forecaster.
        """
        if self.changepoint_interval <= 0:
            raise ValueError("changepoint_interval must be greater than 0.")

        if self.changepoint_prior_scale <= 0:
            raise ValueError("changepoint_prior_scale must be greater than 0.")
        if self.noise_scale <= 0:
            raise ValueError("noise_scale must be greater than 0.")
        if self.capacity_prior_scale <= 0:
            raise ValueError("capacity_prior_scale must be greater than 0.")
        if self.capacity_prior_loc <= 0:
            raise ValueError("capacity_prior_loc must be greater than 0.")

        if self.trend not in ["linear", "logistic"]:
            raise ValueError('trend must be either "linear" or "logistic".')

    def _get_fit_data(self, y, X, fh):
        """
        Prepare the data for the NumPyro model.

        Args:
            y (pd.DataFrame): Training target time series.
            X (pd.DataFrame): Training exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            dict: A dictionary containing the model data.
        """

        # Handling series without __total indexes
        self.aggregator_ = Aggregator()
        self.original_y_indexes_ = y.index
        y = self.aggregator_.fit_transform(y)
        self.full_y_indexes_ = y.index

        # Hierarchy matrix (S Matrix)
        self.hierarchy_matrix = jnp.array(_get_s_matrix(y).values)

        # Updating internal _y of sktime because BaseBayesianForecaster uses it to convert
        # Forecast Horizon into multiindex correcly
        self._y = y

        # Convert inputs to array, including the time index
        y_bottom = loc_bottom_series(y)
        y_bottom_arrays = series_to_tensor(y_bottom)
        t_arrays = extract_timetensor_from_dataframe(y_bottom)

        # Setup time scale
        self._set_time_scale(t_arrays)
        t_scaled = self._time_scaler.scale(t_arrays)
        # We use a single time array for all series, with shape (n_timepoints, 1)
        t_scaled = t_scaled[0].reshape((-1, 1))
        # Changepoints
        self._setup_changepoints(t_scaled=t_scaled)
        changepoint_matrix = self._get_changepoint_matrix(t_scaled)

        # Exog variables

        # If no exogenous variables, create empty DataFrame
        # Else, aggregate exogenous variables and transform them
        if (X is None or X.columns.empty) and self.feature_transformer is not None:
            X = pd.DataFrame(index=y.index)
        if self.feature_transformer is not None:
            X = self.feature_transformer.fit_transform(X)
        self._has_exogenous_variables = X is not None and not X.columns.empty

        if self._has_exogenous_variables:

            self.expand_columns_transformer_ = ExpandColumnPerLevel(
                X.columns.difference(self.shared_features)
            ).fit(X)
            X = X.loc[y.index]
            X = self.expand_columns_transformer_.transform(X)

            self._set_custom_effects(feature_names=X.columns)
            exogenous_data = self._get_exogenous_data_array(loc_bottom_series(X))

        else:
            self._exogenous_effects_and_columns = {}
            exogenous_data = {}

        self.fit_and_predict_data_ = {
            "trend_mode": self.trend,
            "exogenous_effects": self.exogenous_effect_dict,
            "init_trend_params": self._get_trend_sample_func(
                t_arrays=t_scaled, y_arrays=y_bottom_arrays
            ),
            "correlation_matrix_concentration": self.correlation_matrix_concentration,
            "noise_scale": self.noise_scale,
        }

        return dict(
            t=t_scaled,
            y=y_bottom_arrays,
            data=exogenous_data,
            changepoint_matrix=changepoint_matrix,
            **self.fit_and_predict_data_,
        )

    def _get_exogenous_matrix_from_X(self, X: pd.DataFrame) -> jnp.ndarray:
        """
        Convert the exogenous variables to a NumPyro matrix.

        Args:
            X (pd.DataFrame): The exogenous variables.

        Returns:
            jnp.ndarray: The NumPyro matrix of the exogenous variables.
        """
        X_bottom = loc_bottom_series(X)
        X_arrays = series_to_tensor(X_bottom)

        return X_arrays

    def _set_time_scale(self, t_arrays):
        """
        Setup model parameters and scalers.

        This function has the collateral effect of setting the following attributes:
        - self._time_scaler
        - self.max_t
        - self.min_t


        Args:
            t_arrays (ndarray): Transformed time index.

        Returns:
            None
        """

        # Scale time index
        self._time_scaler = TimeScaler()
        self._time_scaler.fit(t=t_arrays)
        t_scaled = self._time_scaler.scale(t=t_arrays)

        # Setting loc and scales for the priors
        self.max_t = t_scaled.max(axis=1).squeeze()
        self.min_t = t_scaled.min(axis=1).squeeze()

    @property
    def _capacity_prior_loc(self):
        """
        The prior location of the capacity parameter.

        Returns:
            float: The prior location of the capacity parameter.
        """

        val = self.capacity_prior_loc
        if isinstance(val, Iterable):
            return jnp.array(val)
        return jnp.ones(self.n_series) * val

    @property
    def _capacity_prior_scale(self):
        """
        The capacity prior scale

        Returns a float or an array-like with the same shape as the number of series.
        """
        val = self.capacity_prior_scale
        if isinstance(val, Iterable):
            return jnp.array(val)
        return jnp.ones(self.n_series) * val

    def _setup_changepoints(self, t_scaled):
        """
        Setup changepoint variables and transformer.

        This function has the collateral effect of setting the following attributes:
        - self._changepoint_ts

        Args:
            t_arrays (ndarray): Transformed time index.

        Returns:
            None
        """
        changepoint_intervals = to_list_if_scalar(
            self.changepoint_interval, self.n_series
        )
        changepoint_ranges = to_list_if_scalar(
            self.changepoint_range or -self.changepoint_interval, self.n_series
        )

        changepoint_ts = []
        for changepoint_interval, changepoint_range in zip(
            changepoint_intervals, changepoint_ranges
        ):
            changepoint_ts.append(
                get_changepoint_timeindexes(
                    t_scaled,
                    changepoint_interval=changepoint_interval,
                    changepoint_range=changepoint_range,
                )
            )

        self._changepoint_ts = changepoint_ts

    def _get_changepoint_matrix(self, t_scaled):
        """
        Get the changepoint matrix. The changepoint matrix has shape (n_series, n_timepoints, total number of changepoints for all series).
        A mask is applied so that for index i at dim 0, only the changepoints for series i are non-zero at dim -1.

        Args:
            t_scaled (ndarray): Transformed time index.

        Returns:
            ndarray: The changepoint matrix.
        """
        changepoint_ts = np.concatenate(self._changepoint_ts)
        changepoint_design_tensor = []
        changepoint_mask_tensor = []
        for i, n_changepoints in enumerate(self.n_changepoint_per_series):
            A = get_changepoint_matrix(t_scaled, changepoint_ts)

            start_idx = sum(self.n_changepoint_per_series[:i])
            end_idx = start_idx + n_changepoints
            mask = np.zeros_like(A)
            mask[:, start_idx:end_idx] = 1

            changepoint_design_tensor.append(A)
            changepoint_mask_tensor.append(mask)

        changepoint_design_tensor = np.stack(changepoint_design_tensor, axis=0)
        changepoint_mask_tensor = np.stack(changepoint_mask_tensor, axis=0)
        return changepoint_design_tensor * changepoint_mask_tensor

    def _get_changepoint_prior_vectors(
        self,
        n_changepoint_per_series: List[int],
        changepoint_prior_scale: float,
        global_rates: jnp.array,
    ):
        """

        Returns the prior vectors for the changepoint coefficients.

        Args:
            n_changepoint_per_series (List[int]): Number of changepoints for each series.
            changepoint_prior_scale (float): Scale parameter for the changepoint prior.
            global_rates (jnp.array): Global rates for each series.

        Returns:
            None
        """

        n_series = len(n_changepoint_per_series)

        def zeros_with_first_value(size, first_value):
            x = jnp.zeros(size)
            x.at[0].set(first_value)
            return x

        changepoint_prior_scale_vector = np.concatenate(
            [
                np.ones(n_changepoint) * cur_changepoint_prior_scale
                for n_changepoint, cur_changepoint_prior_scale in zip(
                    n_changepoint_per_series,
                    to_list_if_scalar(changepoint_prior_scale, n_series),
                )
            ]
        )

        changepoint_prior_loc_vector = np.concatenate(
            [
                zeros_with_first_value(n_changepoint, estimated_global_rate)
                for n_changepoint, estimated_global_rate in zip(
                    n_changepoint_per_series, global_rates
                )
            ]
        )

        return jnp.array(changepoint_prior_loc_vector), jnp.array(
            changepoint_prior_scale_vector
        )

    def _get_trend_sample_func(
        self, t_arrays: jnp.ndarray, y_arrays: jnp.ndarray
    ) -> Callable:
        """

        Get a function that samples the trend parameters.

        This function may change in the future. Currently, the model function receives a function to get the changepoint coefficients.
        This function is passed to the model as a partial function, with the changepoint coefficients as a parameter

        Args:
            t_arrays (jnp.array): array with shape (n_timepoints, 1).
            y_arrays (jnp.array): array with shape (n_series, n_timepoints, 1).

        Returns:
            Callable: Function that samples the trend parameters.

        """

        distributions = self._get_trend_prior_distributions(t_arrays, y_arrays)

        def trend_sample_func(distributions):

            return init_params(distributions)

        return functools.partial(trend_sample_func, distributions=distributions)

    def _get_trend_prior_distributions(self, t_arrays, y_arrays):

        distributions = {}

        if self.trend == "linear":
            global_rates = enforce_array_if_zero_dim(
                (y_arrays[:, -1].squeeze() - y_arrays[:, 0].squeeze())
                / (t_arrays[0].squeeze() - t_arrays[-1].squeeze())
            )
            offset = y_arrays[:, 0].squeeze() - global_rates * t_arrays[0].squeeze()

            distributions["offset"] = dist.Normal(
                offset,
                jnp.array(self.changepoint_prior_scale) * 10,
            )

        if self.trend == "logistic":

            global_rates, offset = suggest_logistic_rate_and_offset(
                t=t_arrays.squeeze(),
                y=y_arrays.squeeze(),
                capacities=self._capacity_prior_loc,
            )
            # We subtract one because it is later added to capacity (inside the model)
            distributions["capacity"] = dist.TransformedDistribution(
                dist.HalfNormal(
                    self._capacity_prior_scale,
                ),
                dist.transforms.AffineTransform(
                    loc=self._capacity_prior_loc, scale=1
                ),
            )

            distributions["offset"] = dist.Normal(
                offset,
                jnp.array(self.changepoint_prior_scale) * 10,
            )

        # Trend

        # Changepoints and trend-related distributions
        changepoint_prior_loc_vector, changepoint_prior_scale_vector = (
            self._get_changepoint_prior_vectors(
                n_changepoint_per_series=self.n_changepoint_per_series,
                changepoint_prior_scale=self.changepoint_prior_scale,
                global_rates=global_rates,
            )
        )

        distributions["changepoint_coefficients"] = dist.Laplace(
            changepoint_prior_loc_vector, changepoint_prior_scale_vector
        )

        return distributions

    def predict_samples(self, X: pd.DataFrame, fh: ForecastingHorizon) -> np.ndarray:
        """Generate samples for the given exogenous variables and forecasting horizon.

        Args:
            X (pd.DataFrame): Exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            np.ndarray: Predicted samples.
        """
        samples = super().predict_samples(X=X, fh=fh)

        return self.aggregator_.transform(samples)

    def _get_predict_data(self, X: pd.DataFrame, fh: ForecastingHorizon) -> np.ndarray:
        """Generate samples for the given exogenous variables and forecasting horizon.

        Args:
            X (pd.DataFrame): Exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            np.ndarray: Predicted samples.
        """

        fh_dates = fh.to_absolute(
            cutoff=self.full_y_indexes_.get_level_values(-1).max()
        )
        fh_as_index = pd.Index(list(fh_dates.to_numpy()))

        if not isinstance(fh, ForecastingHorizon):
            fh = self._check_fh(fh)

        t_arrays = jnp.array(convert_index_to_days_since_epoch(fh_as_index)).reshape(
            (1, -1, 1)
        )
        t_arrays = jnp.tile(t_arrays, (self.n_series, 1, 1))
        t_arrays = self._time_scaler.scale(t_arrays)
        t_scaled = t_arrays[0]

        changepoint_matrix = self._get_changepoint_matrix(t_scaled)

        if self._has_exogenous_variables:
            if X is None or X.shape[1] == 0:
                idx = reindex_time_series(self._y, fh_as_index).index
                X = pd.DataFrame(index=idx)
                X = self.aggregator_.transform(X)

            X = X.loc[X.index.get_level_values(-1).isin(fh_as_index)]
            if self.feature_transformer is not None:
                X = self.feature_transformer.transform(X)
            X = self.expand_columns_transformer_.transform(X)
            exogenous_data = self._get_exogenous_data_array(loc_bottom_series(X))
        else:
            exogenous_data = {}

        return dict(
            t=t_arrays,
            y=None,
            data=exogenous_data,
            changepoint_matrix=changepoint_matrix,
            **self.fit_and_predict_data_,
        )

    def periodindex_to_multiindex(self, periodindex: pd.PeriodIndex) -> pd.MultiIndex:
        """
        Convert a PeriodIndex to a MultiIndex.

        Args:
            periodindex (pd.PeriodIndex): PeriodIndex to convert.

        Returns:
            pd.MultiIndex: Converted MultiIndex.
        """
        if self._y.index.nlevels == 1:
            return periodindex

        levels = self._y.index.droplevel(-1).unique().tolist()
        # import Iterable
        from collections.abc import Iterable

        # Check if base_levels 0 is a iterable, because, if there are more than
        # one level, the objects are tuples. If there's only one level, the
        # objects inside levels list are strings. We do that to make it more "uniform"
        if not isinstance(levels[0], tuple):
            levels = [(x,) for x in levels]

        bottom_levels = [idx for idx in levels if idx[-1] != "__total"]

        return pd.Index(
            map(
                lambda x: (*x[0], x[1]),
                itertools.product(bottom_levels, periodindex),
            ),
            name=self._y.index.names,
        )

    @property
    def n_changepoint_per_series(self):
        """Get the number of changepoints per series.

        Returns:
            int: Number of changepoints per series.
        """
        return [len(cp) for cp in self._changepoint_ts]

    @property
    def n_series(self):
        """Get the number of series.

        Returns:
            int: Number of series.
        """
        return self.hierarchy_matrix.shape[1]


def enforce_array_if_zero_dim(x):
    """
    Reshapes the input array `x` to have at least one dimension if it has zero dimensions.

    Args:
        x (array-like): The input array.

    Returns:
        array-like: The reshaped array.

    """
    if x.ndim == 0:
        return x.reshape(1)
    return x


def to_list_if_scalar(x, size=1):
    """
    Converts a scalar value to a list of the same value repeated `size` times.

    Args:
        x (scalar or array-like): The input value to be converted.
        size (int, optional): The number of times to repeat the value in the list. Default is 1.

    Returns:
        list: A list containing the input value repeated `size` times if `x` is a scalar, otherwise returns `x` unchanged.
    """
    if np.isscalar(x):
        return [x] * size
    return x


class TimeScaler:

    def fit(self, t):
        """
        Fit the time scaler.

        Parameters:
            t (ndarray): Time indices for each series.

        Returns:
            TimeScaler: The fitted TimeScaler object.
        """

        if t.ndim == 1:
            t = t.reshape(1, -1)
        self.t_scale = (t[:, 1:] - t[:, :-1]).flatten().mean()
        self.t_min = t.min()
        return self

    def scale(self, t):
        """
        Transform the time indices.

        Parameters:
            t (ndarray): Time indices for each series.

        Returns:
            ndarray: Transformed time indices.
        """
        return (t - self.t_min) / self.t_scale

    def fit_scale(self, t):
        """
        Fit the time scaler and transform the time indices.

        Parameters:
            t (ndarray): Time indices for each series.

        Returns:
            ndarray: Transformed time indices.
        """
        self.fit(t)
        return self.scale(t)



import numpy as np
import jax.numpy as jnp

def compute_changepoint_design_matrix(t, changepoint_ts):
    """
    Compute the changepoint design matrix.

    Parameters:
        t (ndarray): Time indices for each series.
        changepoint_ts (list): List of changepoint time indices for each series.

    Returns:
        ndarray: Changepoint design matrix.
    """
    t = np.array(t)
    changepoint_ts = [np.array(x) for x in changepoint_ts]

    n_changepoint_per_series = [len(x) for x in changepoint_ts]
    changepoint_ts = np.concatenate(changepoint_ts)
    changepoint_design_tensor = []
    changepoint_mask_tensor = []
    for i, n_changepoints in enumerate(n_changepoint_per_series):
        expanded_ts = jnp.tile(t.reshape((-1, 1)), (1, sum(n_changepoint_per_series)))
        A = (expanded_ts >= changepoint_ts.reshape((1, -1))).astype(int) * expanded_ts
        cutoff_ts = (
            (expanded_ts < changepoint_ts.reshape((1, -1))).astype(int) * expanded_ts
        ).max(axis=0)
        A = np.clip(A - cutoff_ts, 0, None)

        start_idx = sum(n_changepoint_per_series[:i])
        end_idx = start_idx + n_changepoints
        mask = np.zeros_like(A)
        mask[:, start_idx:end_idx] = 1

        changepoint_design_tensor.append(A)
        changepoint_mask_tensor.append(mask)

    changepoint_design_tensor = np.stack(changepoint_design_tensor, axis=0)
    changepoint_mask_tensor = np.stack(changepoint_mask_tensor, axis=0)
    return jnp.array(changepoint_design_tensor) * jnp.array(changepoint_mask_tensor)



""" Univariate Prophet model

This module implements the Univariate Prophet model, similar to the one implemented in the `prophet` library.

"""

from typing import Callable
from functools import partial
from sktime.transformations.series.detrend import Detrender
import jax.numpy as jnp
import pandas as pd
from numpyro import distributions as dist
from jax import random
from sktime.forecasting.base import ForecastingHorizon

from prophetverse.utils.frame_to_array import convert_index_to_days_since_epoch
from prophetverse.sktime.base import (
    BaseBayesianForecaster,
    ExogenousEffectMixin, init_params,
)


from sktime.transformations.base import BaseTransformer
from prophetverse.utils.logistic import suggest_logistic_rate_and_offset

from prophetverse.models.univariate_model import model
from prophetverse.changepoint import (
    get_changepoint_matrix,
    get_changepoint_timeindexes,
)
import functools


__all__ = ["Prophet"]


class Prophet(ExogenousEffectMixin, BaseBayesianForecaster):
    """

    Univariate Prophet model, similar to the one implemented in the `prophet` library.

    The main difference between the mathematical model here and from the original one is the
    logistic trend. Here, another parametrization is considered, and the capacity is not
    passed as input, but inferred from the data.

    With respect to API, this one follows sktime convention where all hiperparameters are passed during
    __init__, and uses `changepoint_interval` instead of `n_changepoints` to set the changepoints. There's no
    weekly_seasonality/yearly_seasonality, but instead, the user can pass a `feature_transformer` that
    will be used to generate the fourier terms. The same for holidays.

    This model accepts configurations where each exogenous variable has a different function relating it to its additive
    effect on the time series. One can, for example, set different priors for a group of feature, or use a Hill function
    to model the effect of a feature.


    Args:
        changepoint_interval (int): The number of points between each potential changepoint.
        changepoint_range (float): Proportion of the history in which trend changepoints will be estimated. 
                                   If a float between 0 and 1, the range will be that proportion of the history. 
                                   If an int, the range will be that number of points. A negative int indicates number of points
                                   counting from the end of the history.
        changepoint_prior_scale (float): Parameter controlling the flexibility of the automatic changepoint selection.
        feature_transformer (BaseTransformer): Transformer object to generate Fourier terms, holiday or other features. Should be a sktime's Transformer
        capacity_prior_scale (float): Scale parameter for the prior distribution of the capacity.
        capacity_prior_loc (float): Location parameter for the prior distribution of the capacity.
        noise_scale (float): Scale parameter for the observation noise.
        trend (str): Type of trend to use. Can be "linear" or "logistic".
        mcmc_samples (int): Number of MCMC samples to draw.
        mcmc_warmup (int): Number of MCMC warmup steps.
        mcmc_chains (int): Number of MCMC chains to run in parallel.
        inference_method (str): Inference method to use. Can be "mcmc" or "map".
        optimizer_name (str): Name of the optimizer to use for variational inference.
        optimizer_kwargs (dict): Additional keyword arguments to pass to the optimizer.
        optimizer_steps (int): Number of optimization steps to perform for variational inference.
        exogenous_effects (List[AbstractEffect]): A list defining the exogenous effects to be used in the model.
        default_effect (AbstractEffect): The default effect to be used when no effect is specified for a variable.
        default_exogenous_prior (tuple): Default prior distribution for exogenous effects.
        rng_key (jax.random.PRNGKey): Random number generator key.


    """

    _tags = {
        "requires-fh-in-fit": False,
        "y_inner_mtype": "pd.DataFrame",
    }

    def __init__(
        self,
        changepoint_interval=25,
        changepoint_range=0.8,
        changepoint_prior_scale=0.001,
        feature_transformer=None,
        capacity_prior_scale=0.2,
        capacity_prior_loc=1.1,
        noise_scale=0.05,
        trend="linear",
        mcmc_samples=2000,
        mcmc_warmup=200,
        mcmc_chains=4,
        inference_method="map",
        optimizer_name="Adam",
        optimizer_kwargs={"step_size" : 1e-4},
        optimizer_steps=100_000,
        exogenous_effects=None,
        default_effect=None,
        rng_key=random.PRNGKey(24),
    ):
        """
        Initializes the Prophet model.
        """

        self.changepoint_interval = changepoint_interval
        self.changepoint_range = changepoint_range
        self.changepoint_prior_scale = changepoint_prior_scale
        self.noise_scale = noise_scale
        self.feature_transformer = feature_transformer
        self.capacity_prior_scale = capacity_prior_scale
        self.capacity_prior_loc = capacity_prior_loc
        self.trend = trend

        super().__init__(
            rng_key=rng_key,
            # ExogenousEffectMixin
            default_effect=default_effect,
            exogenous_effects=exogenous_effects,
            # BaseBayesianForecaster
            inference_method=inference_method,
            mcmc_samples=mcmc_samples,
            mcmc_warmup=mcmc_warmup,
            mcmc_chains=mcmc_chains,
            optimizer_name=optimizer_name,
            optimizer_kwargs=optimizer_kwargs,
            optimizer_steps=optimizer_steps,
        )

        self.model = model
        self._validate_hyperparams()

    def _validate_hyperparams(self):
        """
        Validate the hyperparameters
        """
        if self.changepoint_interval <= 0:
            raise ValueError("changepoint_interval must be greater than 0.")

        if self.changepoint_prior_scale <= 0:
            raise ValueError("changepoint_prior_scale must be greater than 0.")
        if self.noise_scale <= 0:
            raise ValueError("noise_scale must be greater than 0.")
        if self.capacity_prior_scale <= 0:
            raise ValueError("capacity_prior_scale must be greater than 0.")
        if self.capacity_prior_loc <= 0:
            raise ValueError("capacity_prior_loc must be greater than 0.")
        if self.trend not in ["linear", "logistic"]:
            raise ValueError('trend must be either "linear" or "logistic".')

    def _get_fit_data(self, y, X, fh):
        """
        Prepares the data for the Numpyro model.

        Args:
            y (pd.DataFrame): Time series data.
            X (pd.DataFrame): Exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            dict: Dictionary of data for the Numpyro model.
        """

        self._set_time_scale(y)

        ## Changepoints and trend
        self._set_changepoints_t(y)
        t = self._index_to_scaled_timearray(y.index)
        changepoint_matrix = self._get_changepoint_matrix(t)
        trend_sample_func = self._get_trend_sample_func(y=y, X=X)

        ## Exogenous features

        if X is None:
            X = pd.DataFrame(index=y.index)

        if self.feature_transformer is not None:

            X = self.feature_transformer.fit_transform(X)

        self._has_exogenous = ~X.columns.empty
        X = X.loc[y.index]

        self._set_custom_effects(X.columns)
        exogenous_data = self._get_exogenous_data_array(X)

        y_array = jnp.array(y.values.flatten()).reshape((-1, 1))

        ## Inputs that also are used in predict
        self.fit_and_predict_data_ = {
            "init_trend_params": trend_sample_func,
            "trend_mode": self.trend,
            "exogenous_effects": self.exogenous_effect_dict if self._has_exogenous else None,
            }

        inputs = {
            "t": self._index_to_scaled_timearray(y.index),
            "y": y_array,
            "data": exogenous_data,
            "changepoint_matrix": changepoint_matrix,
            **self.fit_and_predict_data_,
        }

        return inputs

    def _get_trend_sample_func(self, y: pd.DataFrame, X: pd.DataFrame) -> Callable :
        """
        
        Get a function that samples the trend parameters.
        
        This function may change in the future. Currently, the model function receives a function to get the changepoint coefficients.
        This function is passed to the model as a partial function, with the changepoint coefficients as a parameter
        
        Args:
            y (pd.DataFrame): Time series data.
            X (pd.DataFrame): Exogenous variables.
            
        Returns:
            Callable: Function that samples the trend parameters.
        
        """
        t_scaled = self._index_to_scaled_timearray(y.index)
        distributions = {}

        changepoints_loc = jnp.zeros(len(self._changepoint_t))
        detrender = Detrender()
        trend = y - detrender.fit_transform(y)

        if self.trend == "linear":

            linear_global_rate = (trend.values[-1, 0] - trend.values[0, 0]) / (
                t_scaled[-1] - t_scaled[0]
            )
            changepoints_loc.at[0].set(linear_global_rate)

            distributions["changepoint_coefficients"] = dist.Laplace(
                changepoints_loc,
                jnp.ones(len(self._changepoint_t)) * (self.changepoint_prior_scale),
            )

            distributions["offset"] = dist.Normal(
                (trend.values[0, 0] - linear_global_rate * t_scaled[0]),
                0.1,
            )

        if self.trend == "logistic":

            linear_global_rate, timeoffset = suggest_logistic_rate_and_offset(
                t_scaled,
                trend.values.flatten(),
                capacities=self.capacity_prior_loc,
            )

            linear_global_rate = linear_global_rate[0]
            timeoffset = timeoffset[0]
            changepoints_loc.at[0].set(linear_global_rate)

            changepoint_coefficients_distribution = dist.Laplace(
                changepoints_loc,
                jnp.ones(len(self._changepoint_t)) * self.changepoint_prior_scale,
            )

            distributions["changepoint_coefficients"] = (
                changepoint_coefficients_distribution
            )

            distributions["offset"] = dist.Normal(timeoffset, jnp.log(2))

            distributions["capacity"] = dist.TransformedDistribution(
                dist.HalfNormal(
                    self.capacity_prior_scale,
                ),
                dist.transforms.AffineTransform(loc=self.capacity_prior_loc, scale=1),
            )

        distributions["std_observation"] = dist.HalfNormal(self.noise_scale)

        def init_trend_params(distributions) -> dict:

            return init_params(distributions)

        return functools.partial(init_trend_params, distributions=distributions)

    def _set_time_scale(self, y: pd.DataFrame):
        """
        Sets the scales for the time series data.

        Args:
            y (pd.DataFrame): Time series data.
        """

        # Set time scale
        t_days = convert_index_to_days_since_epoch(y.index)

        self.t_scale = (t_days[1:] - t_days[:-1]).mean()
        self.t_start = t_days.min() / self.t_scale

    def _index_to_scaled_timearray(self, idx):
        """
        Scales the index values.

        Args:
            idx (pd.Index): Pandas Index object.

        Returns:
            np.ndarray: Scaled index values.
        """
        t_days = convert_index_to_days_since_epoch(idx)
        return (t_days) / self.t_scale - self.t_start

    def _set_changepoints_t(self, y : pd.DataFrame) -> None:
        """
        Sets the array of changepoint times.
        
        This function has the colateral effect of setting the attribute `_changepoint_t` in the class.

        Args:
            y (pd.DataFrame): Time series data.
        """

        t_scaled = self._index_to_scaled_timearray(y.index)

        self._changepoint_t = get_changepoint_timeindexes(
            t=t_scaled,
            changepoint_interval=self.changepoint_interval,
            changepoint_range=self.changepoint_range,
        )

    def _get_changepoint_matrix(self, t: jnp.ndarray) -> jnp.ndarray:
        """
        Generates the changepoint coefficient matrix.

        Args:
            t (jnp.ndarray): Array of time values.

        Returns:
            jnp.ndarray: Changepoint coefficient matrix.
        """

        return get_changepoint_matrix(t, self._changepoint_t)

    def _get_predict_data(
        self, X: pd.DataFrame, fh: ForecastingHorizon
    ) -> dict:
        """
        Prepares the data for making predictions.

        Args:
            X (pd.DataFrame): Exogenous variables.
            fh (ForecastingHorizon): Forecasting horizon.

        Returns:
            dict: Dictionary of data for the Numpyro model.
        """

        fh_dates = self.fh_to_index(fh)
        fh_as_index = pd.Index(list(fh_dates.to_numpy()))

        t = self._index_to_scaled_timearray(fh_as_index)
        changepoint_matrix = self._get_changepoint_matrix(t)

        if X is None:
            X = pd.DataFrame(index=fh_as_index)

        if self.feature_transformer is not None:
            X = self.feature_transformer.transform(X)

        exogenous_data = (
            self._get_exogenous_data_array(X.loc[fh_as_index]) if self._has_exogenous else None
        )

        return dict(
            t=t.reshape((-1, 1)),
            y=None,
            data=exogenous_data,
            changepoint_matrix=changepoint_matrix,
            **self.fit_and_predict_data_,
        )

   


#  pylint: disable=g-import-not-at-top
from typing import Protocol, TypedDict, Dict, Tuple, Callable, List
import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist
import re
import logging
import pandas as pd
from abc import ABC, abstractmethod


# --------------
#     Effects
# --------------

class AbstractEffect(ABC):
    """Abstract class for effects.
    
    Effects should inherit from this class and implement the `compute_effect` method.
    The id is used to create unique names for the samples in the model.
    
    """

    def __init__(self, id="", regex=None, **kwargs):
        self.id = id
        self.regex = regex
        
        
    def match_columns(self, columns : pd.Index) -> pd.Index:
        """Match the columns of the DataFrame with the regex pattern.
        
        Args:
            X (pd.DataFrame): The DataFrame to match.
        
        Returns:
            pd.Index: The columns that match the regex pattern.
        """
        
        if isinstance(columns, List):
            columns = pd.Index(columns)
            
        if self.regex is None:
            raise ValueError("To use this method, you must set the regex pattern")
        return columns[columns.str.match(self.regex)]
    
    @staticmethod
    def split_data_into_effects(X : pd.DataFrame, effects : List) -> Dict[str, pd.DataFrame]:
        """Split the data into effects.
        
        Args:
            X (pd.DataFrame): The DataFrame to split.
            effects (List[AbstractEffect]): The effects to split the data into.
        
        Returns:
            Dict[str, pd.DataFrame]: A dictionary mapping effect names to DataFrames.
        """
        data = {}
        for effect in effects:
            data[effect.id] = X[effect.match_columns(X)]
        return data

    def sample(self, name : str, *args, **kwargs):
        """
        Sample a random variable with a unique name.
        """
        return numpyro.sample(f"{self.id}__{name}", *args, **kwargs)

    @abstractmethod
    def compute_effect(self, trend : jnp.ndarray, data : jnp.ndarray) -> jnp.ndarray: 
        """Compute the effect based on the trend and data.
        
        Args:
            trend (jnp.ndarray): The trend.
            data (jnp.ndarray): The data concerning this effect.
        
        Returns:
            jnp.ndarray: The effect.
        """
        ...

    def __call__(self, trend, data):

        return self.compute_effect(trend, data)


class LogEffect(AbstractEffect):
    """
    Log effect for a variable.

    Computes the effect using the formula:

    effect = scale * log(rate * data + 1)

    A gamma prior is used for the scale and rate parameters.

    Args:
        id (str): The identifier for the effect.
        scale_prior (dist.Distribution): The prior distribution for the scale parameter.
        rate_prior (dist.Distribution): The prior distribution for the rate parameter.
    """

    def __init__(
        self,
        scale_prior=dist.Gamma(1, 1),
        rate_prior=dist.Gamma(1, 1),
        effect_mode="multiplicative",
        **kwargs,
    ):
        self.scale_prior = scale_prior
        self.rate_prior = rate_prior
        self.effect_mode = effect_mode
        super().__init__(**kwargs)

    def compute_effect(self, trend, data):
        """
        Computes the effect using the log transformation.

        Args:
            trend: The trend component.
            data: The input data.

        Returns:
            The computed effect.
        """
        scale = self.sample("log_scale", self.scale_prior)
        rate = self.sample("log_rate", self.rate_prior)
        effect = scale * jnp.log(rate * data + 1)
        if self.effect_mode == "additive":
            return effect
        return trend * effect


class LinearEffect(AbstractEffect):
    """
    Represents a linear effect in a hierarchical prophet model.

    Args:
        id (str): The identifier for the effect.
        prior (tuple): A tuple with the distribution class to use for sampling coefficients and  the arguments to pass to the distribution class. Defaults to (dist.Normal, 0, 1).
        effect_mode (str): The mode of the effect, either "multiplicative" or "additive".

    Attributes:
        dist (type): The distribution class used for sampling coefficients.
        dist_args (tuple): The arguments passed to the distribution class.
        effect_mode (str): The mode of the effect, either "multiplicative" or "additive".

    Methods:
        compute_effect(trend, data): Computes the effect based on the given trend and data.

    """

    def __init__(
        self,
        prior=(dist.Normal, 0, 1),
        effect_mode="multiplicative",
        **kwargs):
        self.prior = prior
        self.effect_mode = effect_mode
        super().__init__(**kwargs)

    def compute_effect(self, trend, data):
        """
        Computes the effect based on the given trend and data.

        Args:
            trend: The trend component of the hierarchical prophet model.
            data: The data used to compute the effect.

        Returns:
            The computed effect based on the given trend and data.

        """
        n_features = data.shape[-1]

        dist = self.prior[0]
        dist_args = self.prior[1:]
        coefficients = self.sample(
            "coefs",
            dist(*[jnp.array([arg] * n_features) for arg in dist_args]),
        )

        if coefficients.ndim == 1:
            coefficients = jnp.expand_dims(coefficients, axis=-1)

        if data.ndim == 3 and coefficients.ndim == 2:
            coefficients = jnp.expand_dims(coefficients, axis=0)
        if self.effect_mode == "multiplicative":
            return multiplicative_effect(trend, data, coefficients)
        return additive_effect(trend, data, coefficients)

class LinearHeterogenousPriorsEffect(AbstractEffect):
    """
    Represents a linear effect with heterogenous priors.

    This effect applies a linear transformation to the input data using coefficients sampled from
    heterogenous priors. The effect can be either multiplicative or additive.

    Args:
        exogenous_priors (Dict[str, Tuple[dist.Distribution, ...]]: A dictionary
            mapping regular expressions to tuples of distribution names and their corresponding
            parameters. These priors are used to sample coefficients for each matching column in
            the input data.
        feature_names (pd.Index): The index of feature names in the input data.
        default_exogenous_prior (Tuple[dist.Distribution, ...], optional): The default prior to use
            for columns that do not match any regular expressions. Defaults to (dist.Normal, 0, 1).
        effect_mode (str, optional): The effect mode, either "multiplicative" or "additive".
            Defaults to "multiplicative".
        id (str, optional): The identifier for the effect. Defaults to "".

    Attributes:
        exogenous_priors (Dict[str, Tuple[dist.Distribution, ...]]): A dictionary
            mapping regular expressions to tuples of distribution names and their corresponding
            parameters.
        feature_names (pd.Index): The index of feature names in the input data.
        default_exogenous_prior (Tuple[dist.Distribution, ...]): The default prior to use for
            columns that do not match any regular expressions.
        effect_mode (str): The effect mode, either "multiplicative" or "additive".
        exogenous_permutation_matrix (ndarray): The permutation matrix used to map coefficients to
            their corresponding columns in the input data.
        exogenous_dists (List[Tuple[str, dist.Distribution]]): A list of tuples containing the name
            of the coefficient and the corresponding distribution.

    """

    def __init__(self,
                 exogenous_priors: Dict[str, Tuple[dist.Distribution, ...]],
                 feature_names: pd.Index,
                 default_exogenous_prior=(dist.Normal, 0, 1),
                 effect_mode="multiplicative",
                 **kwargs):
        self.exogenous_priors = exogenous_priors
        self.feature_names = pd.Index(feature_names)
        self.default_exogenous_prior = default_exogenous_prior
        self.effect_mode = effect_mode
        super().__init__(**kwargs)

        self.set_distributions_and_permutation_matrix()

    def compute_effect(self, trend, data):
        """
        Compute the effect on the trend using the sampled coefficients.

        Args:
            trend: The trend data.
            data: The input data.

        Returns:
            The computed effect.

        """
        coefficients = self.get_coefficients()
        if self.effect_mode == "multiplicative":
            return multiplicative_effect(trend, data, coefficients)
        return additive_effect(trend, data, coefficients)

    @property
    def features_with_default_priors(self):
        """
        Check the regular expressions in the exogenous priors and return the remaining regex.

        This function takes the exogenous priors and the input DataFrame `X`.
        It checks if any columns are already set based on the regular expressions in the exogenous priors.
        It returns the remaining regex that matches the columns that are not already set.

        Returns:
            str: The remaining regex that matches the columns that are not already set.

        Raises:
            ValueError: If any columns are already set.

        """
        already_set_columns = set()
        for regex, _ in self.exogenous_priors.items():
            columns = [column for column in self.feature_names if re.match(regex, column)]
            if already_set_columns.intersection(columns):
                raise ValueError(
                    "Columns {} are already set".format(
                        already_set_columns.intersection(columns)
                    )
                )
            already_set_columns = already_set_columns.union(columns)
        remaining_columns = self.feature_names.difference(already_set_columns)

        # Create a regex that matches all remaining columns
        remaining_regex = "|".join(remaining_columns)
        return remaining_regex

    def set_distributions_and_permutation_matrix(self):
        """
        Set the distributions and permutation matrix based on the exogenous priors and feature names.

        """
        if self.feature_names is None or len(self.feature_names) == 0:
            return [], jnp.array([])

        exogenous_dists = []
        exogenous_permutation_matrix = []
        exogenous_priors = self.exogenous_priors.copy()

        if self.features_with_default_priors:
            exogenous_priors[self.features_with_default_priors] = self.default_exogenous_prior

        for i, (regex, (Distribution, *args)) in enumerate(exogenous_priors.items()):
            # Find columns that match the regex
            columns = [column for column in self.feature_names if re.match(regex, column)]
            # Get idx of columns that match the regex
            idx = jnp.array([self.feature_names.get_loc(column) for column in columns])
            # Set the distribution for each column that matches the regex
            distribution: dist.Distribution = Distribution(
                    *[jnp.ones(len(idx)) * arg for arg in args]
                )

            name = "exogenous_coefficients_{}".format(i)

            if not len(idx):
                logging.warning(
                        "No columns in the DataFrame match the regex pattern: {}".format(regex)
                    )
                continue
            # Matrix of shape (len(columns), len(idx) that map len(idx) to the corresponding indexes
            exogenous_permutation_matrix.append(jnp.eye(len(self.feature_names))[idx].T)
            exogenous_dists.append((name, distribution))

        self.exogenous_permutation_matrix = jnp.concatenate(exogenous_permutation_matrix, axis=1)
        self.exogenous_dists = exogenous_dists

    def get_coefficients(self):
        """
        Get the sampled coefficients based on the exogenous distributions.

        Returns:
            The sampled coefficients.

        """
        parameters = []
        for regex, distribution in self.exogenous_dists:
            parameters.append(
                self.sample(
                    "exogenous_coefficients_{}".format(regex), distribution
                ).reshape((-1, 1))
            )

        return self.exogenous_permutation_matrix @ jnp.concatenate(parameters, axis=0)


class HillEffect(AbstractEffect):
    """
    Represents a Hill effect in a time series model.

    Attributes:
        half_max_prior: Prior distribution for the half-maximum parameter.
        slope_prior: Prior distribution for the slope parameter.
        max_effect_prior: Prior distribution for the maximum effect parameter.
        effect_mode: Mode of the effect (either "additive" or "multiplicative").
    """

    def __init__(
        self,
        half_max_prior=dist.Gamma(1, 1),
        slope_prior=dist.HalfNormal(10),
        max_effect_prior=dist.Gamma(1, 1),
        effect_mode="multiplicative",
        **kwargs,
    ):
        self.half_max_prior = half_max_prior
        self.slope_prior = slope_prior
        self.max_effect_prior = max_effect_prior
        self.effect_mode = effect_mode
        super().__init__(**kwargs)

    def compute_effect(self, trend, data):
        """
        Computes the effect using the log transformation.

        Args:
            trend: The trend component.
            data: The input data.

        Returns:
            The computed effect.
        """

        half_max = self.sample("half_max", self.half_max_prior)
        slope = self.sample("slope", self.slope_prior)
        max_effect = self.sample("max_effect", self.max_effect_prior)

        effect = max_effect * (1 / (1 + (data / half_max) ** -slope))

        if self.effect_mode == "additive":
            return effect
        return trend * effect


def matrix_multiplication(data, coefficients):
    return data @ coefficients.reshape((-1, 1))


def additive_effect(
    trend: jnp.ndarray, data: jnp.ndarray, coefficients: jnp.ndarray
) -> jnp.ndarray:
    return matrix_multiplication(data, coefficients)


def multiplicative_effect(
    trend: jnp.ndarray, data: jnp.ndarray, coefficients: jnp.ndarray
) -> jnp.ndarray:
    return trend * matrix_multiplication(data, coefficients)





def starts_with(prefixes):
    """
    Returns a regular expression pattern that matches strings starting with any of the given prefixes.

    Args:
        prefixes (list): A list of strings representing the prefixes to match.

    Returns:
        str: A regular expression pattern that matches strings starting with any of the given prefixes.
    """
    if isinstance(prefixes, str):
        prefixes = [prefixes]
    return rf"^(?:{'|'.join(prefixes)})"


def exact(string):
    """
    Returns a regular expression pattern that matches the exact given string.

    Args:
        string (str): The string to match exactly.

    Returns:
        str: A regular expression pattern that matches the exact given string.
    """
    return rf"^{string}$"